/* $Workfile: Vd5x16.c $       $Revision: 15 $      $Modtime: 7/28/00 2:58p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    Vd5X16.C

    A simple example console application demonstrating the use of OTX Driver
    with the Vidar-5x16-PCI board.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    The application loads the standard Otx Signal Processing Module 1 DSP
    Application Pack to all DSPs and configures DSP#0 to send a sine
    wave to the handset connected to Codec#0.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"

#define DEMO_TASK_REF  1L               // Task Reference for a non-blocking call
#define DEMO_DSP_MAX   16               // This demo is designed for full 16 DSPs
#define DEMO_CODEC_MAX  1               // Use only codec

#define DEMO_DEFAULT_LOG_FILE "Vd5x16.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char           *logFileName;
  OtxDataFormatE  law;
  int             boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
char				g_sqNumber0[] = "19726640100";
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
static      OTX_BOOL g_bSineWaveOn[DEMO_DSP_MAX];

struct {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hCodec[DEMO_CODEC_MAX];
  OTX_HANDLE hDsp[DEMO_DSP_MAX];
} g_hPhysDevs;


struct {
  OTX_HANDLE hDspCodec[DEMO_DSP_MAX];   // Pipe DSP --> Codec
  OTX_HANDLE hCodecDsp[DEMO_DSP_MAX];   // Pipe Codec --> DSP
  OTX_HANDLE hDspDsp[DEMO_DSP_MAX];     // Pipe DSP --> DSP (Dtmf Dialler --> Dtmf Detector)
} g_hPipes;


struct {
  OTX_HANDLE hToneGenerator[DEMO_DSP_MAX];
  OTX_HANDLE hDtmfDetector[DEMO_DSP_MAX];
  OTX_HANDLE hDtmfDialler[DEMO_DSP_MAX];
} g_hLogicDevs;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (result == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Vidar-5x16-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;

  // Connect to the Physical Vidar-5x16-PCI Board Controller
  result = OtxDrvOpenPhysicalDevice(0, OTX_DEVICE_VIDAR_5x_PCI, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Connect to DSP Devices
  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  }

  // Get handles to Codec device #0
  for (i = 0; i < DEMO_CODEC_MAX; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  }

  // Load DSP Program
  for (nDsp = 0; nDsp < DEMO_DSP_MAX; nDsp++) {
    DemoLogf("Loading program OTX_DSP_SPM_OTXSPM1 to DSP#%d\n", nDsp);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], OTX_DSP_SPM_OTXSPM1); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  }

}


/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;


  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }


}

/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  nDsp;


  for (nDsp = 0; nDsp < DEMO_DSP_MAX; nDsp++) {
    // Create Tone Effect Generators Logical Devices (one per DSP)
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_EFFECTS_GENERATOR, g_hEventQueue, nDsp, &g_hLogicDevs.hToneGenerator[nDsp]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);

    // Create Dtmf Detector Logical Devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_DTMF_DETECTOR, g_hEventQueue, nDsp, &g_hLogicDevs.hDtmfDetector[nDsp]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);

    // Create DTMF Dialers Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_DTMF_DIALER, g_hEventQueue, nDsp, &g_hLogicDevs.hDtmfDialler[nDsp]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset
    of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a
    characterized by its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  // Create Pipes
  for (i = 0; i < DEMO_DSP_MAX; i++) {
    // Dsp <--> Codec
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspCodec[i]); DemoCheckRc(g_hPipes.hDspCodec[i], result);  // DSP --> CODEC
    result = OtxDrvCreatePipe(64, &g_hPipes.hCodecDsp[i]); DemoCheckRc(g_hPipes.hCodecDsp[i], result);  // CODEC --> DSP
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspDsp[i]); DemoCheckRc(g_hPipes.hDspDsp[i], result);  // DSP --> DSP

  }

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    if (i < 8) {
      // DSP 0-7 all get a full highway (32 Timeslots). Let the driver decided which 
      // timeslot to use (supply NULL as the timeSlotMask).

      // Connect one end of the handset Pipes to DSPs
      result = OtxDrvConnectPipe(g_hPipes.hDspCodec[i], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[i], result);
      // Connect both ends of the Dtmf Dialler --> Dtmf Detector to the same DSP
      result = OtxDrvConnectPipe(g_hPipes.hDspDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspDsp[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hDspDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspDsp[i], result);
    } else {
      // DSP 8-11  share one highway (8) with 8 timeslots each
      // DSP 8  gets TS 0-7
      // DSP 9  gets TS 8-15
      // DSP 10 gets TS 16-23
      // DSP 11 gets TS 24-31
      // DSP 12-15 share one highway (9) with 8 timeslots each
      // DSP 12 gets TS 0-7
      // DSP 13 gets TS 8-15
      // DSP 14 gets TS 16-23
      // DSP 15 gets TS 24-31
      ts = (i%4) * 8; 
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;

      // Connect one end of the handset Pipes to DSPs
      result = OtxDrvConnectPipe(g_hPipes.hDspCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspCodec[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hCodecDsp[i], result);

      ts = (i%4) * 8 + 1; 
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;

      // Connect both ends of the Dtmf Dialler --> Dtmf Detector to the same DSP
      result = OtxDrvConnectPipe(g_hPipes.hDspDsp[i], OTX_PIPE_OUTPUT,   g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspDsp[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hDspDsp[i], OTX_PIPE_INPUT,    g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspDsp[i], result);
    }
  }

  // Connect logical devices within the DSPs
  for (i = 0; i < DEMO_DSP_MAX; i++) {
    // transmit direction: Tone Generator --> Codec
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hToneGenerator[i], -1, g_hPipes.hDspCodec[i], -1); DemoCheckRc(g_hLogicDevs.hToneGenerator[i], result);

    // Dtmf Dialler --> Dtmf Detector
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hDtmfDialler[i], -1, g_hPipes.hDspDsp[i], -1); DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hDspDsp[i], -1, g_hLogicDevs.hDtmfDetector[i], -1); DemoCheckRc(g_hPipes.hDspDsp[i], result);

  }

  // Initially Connect DSP#0 to Codec#0 (Switch the connection with '+' and '-' keys
  result = OtxDrvConnectPipe(g_hPipes.hDspCodec[0], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[0], result);
  result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[0], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[0], result);

}


/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_DSP_MAX; i++) {

    // Initialize the Tone Effect Generators
    result = OtxToneEffectsGeneratorInit(g_hLogicDevs.hToneGenerator[i], OTX_DATA_FORMAT_COMPAND8_U_LAW, /*OTX_TONE_SWEEPING_SINE*/OTX_TONE_SINEWAVE, OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hToneGenerator[i], result);
    g_bSineWaveOn[i] = OTX_FALSE;

    // Initialize the DTMF Diallers
    result = OtxToneDtmfDiallerInit(g_hLogicDevs.hDtmfDialler[i], OTX_DATA_FORMAT_COMPAND8_U_LAW, 12, 12, OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);

    // Initialize the DTMF Detectors
    result = OtxToneDtmfDetectorInit(g_hLogicDevs.hDtmfDetector[i], OTX_DATA_FORMAT_COMPAND8_U_LAW, OTX_TRUE); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);
  }
}

/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hDtmfDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDetector[i], result);
    result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDtmfDialler[i], result);
  }

}




static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OtxDrvReleaseLib();
      OsaConsoleExit();
			exit(1);
    }

  }
}



/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
static void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_HANDLE     hCurrentDsp = 0;
  static OTX_INT32      nCurrentDsp = 0;
  static OTX_BOOL g_bSineWaveOn[DEMO_DSP_MAX];
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" t   Toggle Tone generation On/Off using the current DSP\n");
    DemoLogf(" d   Dial digits '%s' using the current DSP (digits should be detected by the DMTF detector)\n", g_sqNumber0);
    DemoLogf(" +   Select the next DSP # as active DSP (target of subsequent commands)\n");
    DemoLogf(" -   Select the previous DSP # as the active DSP (target of subsequent commands)\n");
    break;

  case '-':
    {
      OTX_INT32 nPrevDsp = nCurrentDsp;

      if ( nCurrentDsp == 0 ) {
        nCurrentDsp = DEMO_DSP_MAX - 1;
      } else {
        nCurrentDsp--;
      }

      // Disconnect Codec#0 from the old active DSP
      result = OtxDrvDisconnectPipe(g_hPipes.hDspCodec[nPrevDsp], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hDspCodec[nPrevDsp], result);
      result = OtxDrvDisconnectPipe(g_hPipes.hCodecDsp[nPrevDsp], OTX_PIPE_INPUT);  DemoCheckRc(g_hPipes.hCodecDsp[nPrevDsp], result);

      // Reconnect Codec #0 to the current active DSP
      result = OtxDrvConnectPipe(g_hPipes.hDspCodec[nCurrentDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[nCurrentDsp], result);
      result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[nCurrentDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[nCurrentDsp], result);

      hCurrentDsp = g_hPhysDevs.hDsp[nCurrentDsp];
      DemoLogf("DSP#%d Selected as the Active DSP ....\n", nCurrentDsp);
    }
    break;

  case '+':
    {
      OTX_INT32 nPrevDsp = nCurrentDsp;

      if ( nCurrentDsp == (DEMO_DSP_MAX - 1) ) {
        nCurrentDsp = 0;
      } else {
        nCurrentDsp++;
      }

      // Disconnect Codec#0 from the previously active DSP
      result = OtxDrvDisconnectPipe(g_hPipes.hDspCodec[nPrevDsp], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hDspCodec[nPrevDsp], result);
      result = OtxDrvDisconnectPipe(g_hPipes.hCodecDsp[nPrevDsp], OTX_PIPE_INPUT);  DemoCheckRc(g_hPipes.hCodecDsp[nPrevDsp], result);

      // Reconnect Codec #0 to the current active DSP
      result = OtxDrvConnectPipe(g_hPipes.hDspCodec[nCurrentDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[nCurrentDsp], result);
      result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[nCurrentDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[nCurrentDsp], result);

      hCurrentDsp = g_hPhysDevs.hDsp[nCurrentDsp];
      DemoLogf("DSP#%d Selected as the Active Dsp ....\n", nCurrentDsp);
    }
    break;

  case 'd':
    {
      OTX_INT32  nDigits = strlen(g_sqNumber0);
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentDsp];

      DemoLogf("Dialling number '%s' on current Dsp#%d \n", g_sqNumber0, nCurrentDsp);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxToneDtmfDial(hHandle, (OTX_UINT8 *)g_sqNumber0, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 't':
    {
      OTX_HANDLE hHandle;
      OTX_BOOL   *pbOn;

      pbOn = &g_bSineWaveOn[nCurrentDsp];
      hHandle = g_hLogicDevs.hToneGenerator[nCurrentDsp];

      if (!*pbOn) {
        // Turn it On
        // Turn Sine Wawe On using the Tone Effects Generator
        DemoLogf("Turn on Sine Wave On on Line #%d: \n", nCurrentDsp);

        // Send DTMF Digits, Blocking call, wait until completed
        result = OtxDrvEnable(hHandle, 0L, 5000L); DemoCheckRc(hHandle, result);
        *pbOn = OTX_TRUE;

      } else {
        // Turn it Off
        // Turn Sine Wawe On using the Tone Effects Generator
        DemoLogf("Turn off Sine Wave On on Line #%d: \n", nCurrentDsp);

        // Send DTMF Digits, Blocking call, wait until completed
        result = OtxDrvDisable(hHandle, 0L, 5000L); DemoCheckRc(hHandle, result);

        *pbOn = OTX_FALSE;
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName  = DEMO_DEFAULT_LOG_FILE;
  envP->law          = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo      = DEMO_DEFAULT_BOARD_NO;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'C') {   // Companding Law
        if (*(argv[i]+2) == 'A') {   // A-law
          envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
        } else if (*(argv[i]+2) == 'U') {   // U-law
          envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
        } else {
          DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
					exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("Vd5x16, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Vidar-5x16-PCI Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("\nPress 'h' for help\n");

  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  DemoLogf("Lib connected okay\n");

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Vidar-5x16-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Enable Physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling Physical Devices...\n");
  DemoEnablePhysicalDevices();
  DemoLogf("Physical Devices Enabled...\n");

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting Devices...\n");
  DemoConnectDevices();
  DemoLogf("Devices connected\n");

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing Logical Devices...\n");
  DemoInitializeLogicalDevices();
  DemoLogf("Logical Devices Initialized...\n");

  // Enable Logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling Logical Devices...\n");
  DemoEnableLogicalDevices();
  DemoLogf("Logical Devices Enabled...\n");

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  // Read CPLD Build Numbers from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

	
  DemoLogf("Main Loop Running; press 'h' for command help, 'q' to quit\n");
  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

	DemoLogf("Normal Program Termination\n");

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();


  return 0;
}

